#!/bin/sh

# Shell script for invoking the wing IDE.  If the first argument is --winghome
# then the directory wing is installed in is echoed to stdout and the script exits.
# Other arguments are:
#  --verbose     Print debugging output
#  --silent      Run silently
#  --system-gtk  Use system-wide version of gtk, rather than Wing's copy
#  --private-gtk Use Wing's copy of gtk rather than the system version
#
# To develop and run from sources, set WINGHOME and WINGPYTHON
# 
# Copyright (c) 2000-2007, Archaeopteryx Software, Inc.  All rights reserved.
#
# Written by Stephan R.A. Deibel and John P. Ehresman

# On Darwin only, start X11 and set DISPLAY default to :0
if [ "`uname`" = "Darwin" ] && [ "${WINGIDE_USE_QT4}" = "" ]; then
	if [ "${DISPLAY}" = "" -o "${DISPLAY}" = ":0.0" -o "${DISPLAY}" = ":0" ]; then
		SECONDS=0
		USER=`whoami | cut -c 1-8`
		DISPLAY_DIR=/tmp/.X11-unix
		X11_APP_NAME=X11
		FIRST_TIME=yes
	
		LISTING=`ls -l $DISPLAY_DIR | grep $USER`
		while [ $SECONDS -le 30 ] && [ -z "$LISTING" ]
		do
		   if [ "${FIRST_TIME}" == "yes" ]; then
				FIRST_TIME="no"
				/usr/bin/open -a $X11_APP_NAME
				if [ "$?" != "0" ]; then
					exit 1
				fi
		   fi 	       
		   sleep 1
		   let "SECONDS = $SECONDS + 1"
		   LISTING=`ls -l $DISPLAY_DIR | grep $USER`
		done
		if [ -z "$LISTING" ]; then
			exit 1
		fi

		DISPLAY=:0
		export DISPLAY
	fi
fi

# Find WINGHOME unless caller has passed it in
if [ "${WINGHOME}" = "" ]; then
    WINGHOME=`dirname "$0"`; export WINGHOME
fi

# Check if user requested debug output or we're running from source in which
# case we also want debug output (but can use --silent to force squelching of
# output also when running from source)
if [ "$1" = "--verbose" ]; then
  shift
  DEBUGARG=
elif [ "$1" = "--silent" ]; then
  shift
  DEBUGARG=--squelch-output
else
  DEBUGARG=--squelch-output
fi

# Location of Python interpreter to use, if not already set
if [ "${WINGPYTHON}" = "" ]; then
  if [ -x "${WINGHOME}/bin/PyCore/python" ]; then
      WINGPYTHON="${WINGHOME}/bin/PyCore/python"
  elif [ "${WINGIDE_USE_QT4}" = "1" ]; then
      WINGPYTHON="${WINGHOME}/bin/runtime-python2.7/bin/python2.7"
  else
      WINGPYTHON="python2.5"
  fi
fi

# Just echo home and exit if requested
if [ "$1" = "--winghome" ]; then
    echo ${WINGHOME}
    exit 0
fi

# Where we install our private copy of GTK and related libs
WING_GTK_PREFIX="${WINGHOME}/bin/gtk-bin"

# Disable Ubuntu Unity menu (aka AppMenu) since it does not work w/ Wing
unset UBUNTU_MENUPROXY

if [ "${WINGIDE_USE_QT4}" != "1" ]; then

	# Force use of system-wide or private if requested on command line
	if [ "$1" = "--system-gtk" ]; then
		shift
		WING_USE_SYSTEM_GTK="1"; export WING_USE_SYSTEM_GTK
	elif [ "$1" = "--private-gtk" ]; then
		shift
		WING_USE_SYSTEM_GTK="0"; export WING_USE_SYSTEM_GTK
	
	# Otherwise use system-wide gtk if ~/.wingide4/use-system-gtk is
	# present unless this is OS X.  On OS X, you must use --system-gtk argument
	elif [ "`uname`" != "Darwin" -a -f "${HOME}/.wingide4/use-system-gtk" ]; then
		WING_USE_SYSTEM_GTK="1"; export WING_USE_SYSTEM_GTK
	# Use system gtk if no private gtk
	elif [ ! -d "${WING_GTK_PREFIX}" ]; then
		WING_USE_SYSTEM_GTK="1"; export WING_USE_SYSTEM_GTK
	else
		WING_USE_SYSTEM_GTK="0"; export WING_USE_SYSTEM_GTK
	fi
	WING_ORIG_WING_USE_SYSTEM_GTK=__undefined__; export WING_ORIG_WING_USE_SYSTEM_GTK

	# Use our private gtk installation
	if [ "${WING_USE_SYSTEM_GTK}" = "0" ]; then
	
		# If GTK_PATH and other envs are defined (as on SuSE) we need to undefine them
		# to avoid polluting our private gtk installation w/ system-wide modules
		if [ "${GTK_PATH}" != "" ]; then
			export WING_ORIG_GTK_PATH=${GTK_PATH}
			unset GTK_PATH
		fi
		if [ "${GTK_EXE_PREFIX}" != "" ]; then
			export WING_ORIG_GTK_EXE_PREFIX=${GTK_EXE_PREFIX}
			unset GTK_EXE_PREFIX
		fi
		if [ "${GTK_DATA_PREFIX}" != "" ]; then
			export WING_ORIG_GTK_DATA_PREFIX=${GTK_DATA_PREFIX}
			unset GTK_DATA_PREFIX
		fi
		if [ "${GTK_MODULES}" != "" ]; then
			export WING_ORIG_GTK_MODULES=${GTK_MODULES}
			unset GTK_MODULES
		fi
		if [ "${GTK2_RC_FILES}" != "" ]; then
			export WING_ORIG_GTK2_RC_FILES=${GTK2_RC_FILES}
			unset GTK2_RC_FILES
		fi
		if [ "${PANGO_RC_FILE}" != "" ]; then
			export WING_ORIG_PANGO_RC_FILE=${PANGO_RC_FILE}
			unset PANGO_RC_FILE
		fi
		if [ "${PANGO_EXE_PREFIX}" != "" ]; then
			export WING_ORIG_PANGO_EXE_PREFIX=${PANGO_EXE_PREFIX}
			unset PANGO_EXE_PREFIX
		fi
		if [ "${PANGO_DATA_PREFIX}" != "" ]; then
			export WING_ORIG_PANGO_DATA_PREFIX=${PANGO_DATA_PREFIX}
			unset PANGO_DATA_PREFIX
		fi
		if [ "${XDG_DATA_DIRS}" != "" ]; then
			export WING_ORIG_XDG_DATA_DIRS=${XDG_DATA_DIRS}
			unset XDG_DATA_DIRS
		fi
		if [ "${XDG_DATA_HOME}" != "" ]; then
			export WING_ORIG_XDG_DATA_HOME=${XDG_DATA_HOME}
			unset XDG_DATA_HOME
		fi
		
		# Need to add our private library areas to the library path
		if [ "`uname`" = "Darwin" ]; then
			if [ "$DYLD_LIBRARY_PATH" != "" ]; then
			  WING_ORIG_DYLD_LIBRARY_PATH=${DYLD_LIBRARY_PATH}
			  DYLD_LIBRARY_PATH="${WING_GTK_PREFIX}/lib:${DYLD_LIBRARY_PATH}"
			else
			  WING_ORIG_DYLD_LIBRARY_PATH=__undefined__
			  DYLD_LIBRARY_PATH="${WING_GTK_PREFIX}/lib"
			fi
			export WING_ORIG_DYLD_LIBRARY_PATH
			export DYLD_LIBRARY_PATH
			
			if [ "$DYLD_FALLBACK_LIBRARY_PATH" != "" ]; then
			  WING_ORIG_DYLD_FALLBACK_LIBRARY_PATH=${DYLD_FALLBACK_LIBRARY_PATH}
			else
			  WING_ORIG_DYLD_FALLBACK_LIBRARY_PATH=__undefined__
			fi
			export WING_ORIG_DYLD_FALLBACK_LIBRARY_PATH
			unset DYLD_FALLBACK_LIBRARY_PATH
			
			if [ "$DYLD_INSERT_LIBRARIES" != "" ]; then
			  WING_ORIG_DYLD_INSERT_LIBRARIES=${DYLD_INSERT_LIBRARIES}
			else
			  WING_ORIG_DYLD_INSERT_LIBRARIES=__undefined__
			fi
			export WING_ORIG_DYLD_INSERT_LIBRARIES
			unset DYLD_INSERT_LIBRARIES
			
			if [ "$DYLD_FORCE_FLAT_NAMESPACE" != "" ]; then
			  WING_ORIG_DYLD_FORCE_FLAT_NAMESPACE=${DYLD_FORCE_FLAT_NAMESPACE}
			else
			  WING_ORIG_DYLD_FORCE_FLAT_NAMESPACE=__undefined__
			fi
			export WING_ORIG_DYLD_FORCE_FLAT_NAMESPACE
			unset DYLD_FORCE_FLAT_NAMESPACE
		else
			if [ "$LD_LIBRARY_PATH" != "" ]; then
			  WING_ORIG_LD_LIBRARY_PATH=${LD_LIBRARY_PATH}
			  LD_LIBRARY_PATH="${WING_GTK_PREFIX}/lib:${LD_LIBRARY_PATH}"
			else
			  WING_ORIG_LD_LIBRARY_PATH=__undefined__
			  LD_LIBRARY_PATH="${WING_GTK_PREFIX}/lib"
			fi
			export WING_ORIG_LD_LIBRARY_PATH
			export LD_LIBRARY_PATH
		fi
	fi
	
	# If display is not set, print an error message
	if [ "${DISPLAY}" = "" ]; then
		echo "wing: Error: DISPLAY is not set"
		exit 1
	fi
fi

# Check for psyco
if [ "$1" = "--with-psyco" ]; then
  PYTHON_ARGS="-OE"
else
  PYTHON_ARGS="-OSE"
fi

# Run the IDE
if [ -f "${WINGHOME}/bin/wing.py" ]; then
    exec "${WINGPYTHON}" $PYTHON_ARGS "${WINGHOME}/bin/wing.py" $DEBUGARG "$@"
elif [ -f "${WINGHOME}/src/wing.py" ]; then
    exec "${WINGPYTHON}" $PYTHON_ARGS "${WINGHOME}/src/wing.py" $DEBUGARG "$@"
elif [ "$DEBUGARG" = "" ]; then
    echo "wing: Error: wing.py not found.  Please check your Wing IDE installation"
fi
